"use strict";

interface DomainJson {
    name: string,
    id: string,
}

interface DomainServerJson {
    host: string,
    id: string,
    priority: number,
    weight: number,
    protocols: {
        https: number | undefined,
        wss: number | undefined,
        http: number | undefined,
        ws: number | undefined,
        udp: number | undefined
    }
}

interface OutputEnvelopeJson {
    status: string,
    request_nr: number,
    data: any,
    error: {
        code: string,
        message: string | null,
        description: string | null,
    } | null,
}

interface EventJson {
    data: {
        message: string,
    }
    id: string,
}

interface WellKnownJson {
    domain: DomainJson,
    domain_servers: DomainServerJson[]
}

export class Domain {
    name: string;
    id: string;
    servers: DomainServer[];
    invalidServers: DomainServer[];

    toString() {
        return `[${this.name}]`;
    }

    constructor(jsonData: DomainJson, domainServers: DomainServerJson[]) {
        // FIXME check values
        this.name = jsonData.name;
        this.id = jsonData.id;
        this.servers = [];
        for (const domainServer of domainServers) {
            this.servers.push(new DomainServer(domainServer));
        }
        this.invalidServers = [];
    }

    static async fromName(domainName: string): Promise<Domain> {
        const controller = new AbortController();
        const timer = setTimeout(() => controller.abort(), 2000);

        const response = await fetch(`https://${domainName}/.well-known/usimp.json`, {
            redirect: "manual",
            signal: controller.signal,
        });

        clearTimeout(timer);
        if (!response.ok) {
            throw new Error("Invalid response");
        }

        const data: WellKnownJson = await response.json();
        return new Domain(data.domain, data.domain_servers);
    }

    chooseDomainServer(): DomainServer {
        if (this.servers.length === 0) throw new Error("No domain servers specified");

        const servers = this.servers.filter(srv => !this.invalidServers.map(srv => srv.id).includes(srv.id));
        if (servers.length === 0) throw new Error("No domain servers reachable");
        if (servers.length === 1 && servers[0]) return servers[0];

        const priority = servers.reduce((min, srv) => Math.min(min, srv.priority), Infinity);
        const domainServers = servers.filter(srv => srv.priority === priority);
        if (domainServers.length === 1 && servers[0]) return servers[0];

        const totalWeight = domainServers.reduce((total, srv) => total + srv.weight, 0);
        const w = Math.floor(Math.random() * totalWeight);

        let accumulator = 0;
        for (const srv of domainServers) {
            accumulator += srv.weight;
            if (w < accumulator) return srv;
        }

        throw new Error("Domain server selection did not work correctly");
    }
}

export class DomainServer {
    host: string;
    id: string;
    priority: number;
    weight: number;
    protocols: {
        https: number | undefined,
        wss: number | undefined,
        http: number | undefined,
        ws: number | undefined,
        udp: number | undefined
    };

    toString() {
        return `[${this.host}/${this.priority}/${this.weight}]`;
    }

    constructor(jsonData: DomainServerJson) {
        this.host = jsonData.host;
        this.id = jsonData.id;
        this.priority = jsonData.priority;
        this.weight = jsonData.weight;
        this.protocols = jsonData.protocols;
    }
}

export class Account {

}

export class Member {

}

export class Room {

}

export class Event {

}

export class Session {
    domain: Domain;
    server: DomainServer | null;
    token: string | null;

    httpBaseUrl: string | null;
    websocket: WebSocket | null;
    numRequests: number;
    subscriptions: string[];
    subscriptionEndpoints: {cb: (a: EventJson) => void}[];

    constructor(domain: Domain) {
        this.domain = domain;
        this.numRequests = 0;
        this.server = null;
        this.token = null;
        this.httpBaseUrl = null;
        this.websocket = null;
        this.subscriptions = [];
        this.subscriptionEndpoints = [];
    }

    close() {
        if (this.websocket && (this.websocket.readyState !== WebSocket.CLOSING && this.websocket.readyState !== WebSocket.CLOSED)) {
            this.websocket.close();
            this.subscriptions = [];
            this.websocket = null;
            this.server = null;
        }
    }

    sleep() {
        this.close();
    }

    async wakeup() {
        await this.chooseDomainServer();
        await this.ping();
        for (const endpoint of this.subscriptionEndpoints) {
            await this._subscribe(endpoint.cb);
        }
    }

    async chooseDomainServer(): Promise<{ http: number | null, ws: number | null } | undefined> {
        while (!this.server) {
            this.server = this.domain.chooseDomainServer();

            const host = this.server.host;
            const protocols = this.server.protocols;

            if ("wss" in protocols) {
                this.websocket = new WebSocket(`wss://${host}:${protocols.wss}/_usimp/websocket`, ["usimp"]);
                this.websocket.addEventListener("error", (error) => {
                    console.error(error);
                    this.websocket = null;
                });
            }

            if ("https" in protocols) {
                this.httpBaseUrl = `https://${host}:${protocols.https}/_usimp`;
                try {
                    return await this.ping();
                } catch {
                    if (this.server !== null) {
                        this.domain.invalidServers.push(this.server);
                        this.server = null;
                    }
                }
            } else {
                console.warn(`Domain server ${this.server} does not support 'https' transport protocol`);
                this.domain.invalidServers.push(this.server);
                this.server = null;
            }
        }
        return undefined;
    }

    waitForWebSocket(timeout: number) {
        if (this.websocket === null) throw new Error("websocket not initialized");
        return new Promise<void>((resolve, reject) => {
            if (this.websocket?.readyState === this.websocket?.OPEN) {
                resolve();
            } else {
                const handlerOpen = () => {
                    clearTimeout(timer);
                    this.websocket?.removeEventListener("open", handlerOpen);
                    this.websocket?.removeEventListener("error", handlerError);
                    resolve();
                };
                const handlerError = (error: any) => {
                    clearTimeout(timer);
                    this.websocket?.removeEventListener("open", handlerOpen);
                    this.websocket?.removeEventListener("error", handlerError);
                    reject(error);
                }
                const handlerTimeout = () => {
                    this.websocket?.removeEventListener("open", handlerOpen);
                    this.websocket?.removeEventListener("error", handlerError);
                    reject(new Error("timeout"));
                }
                const timer = setTimeout(handlerTimeout, timeout);
                this.websocket?.addEventListener("open", handlerOpen);
                this.websocket?.addEventListener("error", handlerError);
            }
        });
    }

    waitForWebSocketResponse(req_nr: number, timeout: number) {
        if (this.websocket === null) throw new Error("websocket not initialized");
        return new Promise<any>((resolve, reject) => {
            const handlerMsg = (msg: any) => {
                const data = JSON.parse(msg.data);
                if (data['request_nr'] === req_nr) {
                    clearTimeout(timer);
                    this.websocket?.removeEventListener("message", handlerMsg);
                    this.websocket?.removeEventListener("error", handlerError);
                    resolve(data);
                }
            }
            const handlerError = (error: any) => {
                clearTimeout(timer);
                this.websocket?.removeEventListener("message", handlerMsg);
                this.websocket?.removeEventListener("error", handlerError);
                reject(error);
            }
            const handlerTimeout = () => {
                this.websocket?.removeEventListener("message", handlerMsg);
                this.websocket?.removeEventListener("error", handlerError);
                reject(new Error("timeout"));
            }
            const timer = setTimeout(handlerTimeout, timeout);
            this.websocket?.addEventListener("message", handlerMsg);
            this.websocket?.addEventListener("error", handlerError);
        });
    }

    async send(endpoint: string, data: any, timeout: number = 2000, forceHttp: boolean = false, cb: ((a: OutputEnvelopeJson) => void) | null = null) {
        this.numRequests++;
        while (true) {
            try {
                if (!forceHttp && this.websocket !== null) {
                    const req_nr = this.numRequests;
                    const startTime = performance.now();

                    await this.waitForWebSocket(timeout);

                    let response;
                    if (cb === null) {
                        response = this.waitForWebSocketResponse(req_nr, timeout);
                    } else {
                        this.websocket.addEventListener("message", (msg) => {
                            const data = JSON.parse(msg.data);
                            if (data['request_nr'] === req_nr) {
                                cb(data);
                            }
                        });
                    }

                    await this.websocket.send(JSON.stringify({
                        'request_nr': req_nr,
                        'endpoint': endpoint,
                        'to_domain': this.domain.id,
                        'token': this.token,
                        'data': data
                    }));

                    if (cb === null) {
                        const responseData = await response;
                        const endTime = performance.now();
                        responseData.duration = endTime - startTime;
                        return responseData;
                    } else {
                        return await this.waitForWebSocketResponse(req_nr, timeout);
                    }
                } else if (this.httpBaseUrl) {
                    const controller = new AbortController();
                    const timer = setTimeout(() => controller.abort(), timeout);

                    let headers: Record<string, string> = {
                        'Content-Type': 'application/json',
                        'To-Domain': this.domain.id,
                    };
                    if (this.token) {
                        headers['Authorization'] = `usimp ${this.token}`;
                    }

                    const startTime = performance.now();
                    const response = await fetch(`${this.httpBaseUrl}/${endpoint}`, {
                        method: "POST",
                        headers: headers,
                        body: JSON.stringify(data),
                        signal: controller.signal,
                    });
                    clearTimeout(timer);
                    const responseData = await response.json();

                    const endTime = performance.now();
                    responseData.duration = endTime - startTime;
                    return responseData;
                } else {
                    throw new Error("No domain server chosen");
                }
            } catch (error) {
                console.error(error);
                if (this.websocket !== null) {
                    this.websocket.close();
                    this.websocket = null;
                }
                this.httpBaseUrl = null;
                if (this.server !== null) {
                    this.domain.invalidServers.push(this.server);
                    this.server = null;
                }
                await this.chooseDomainServer();
            }
        }
    }

    async ping() {
        let result = {"http": null, "ws": null};
        const resHttp = await this.send("ping", {}, undefined, true);
        result.http = resHttp.duration;

        if (this.websocket) {
            const resWs = await this.send("ping", {});
            result.ws = resWs.duration;
        }

        return result;
    }

    async authenticate(accountName: string, password: string) {
        const response = await this.send("authenticate", {
            type: "password",
            name: accountName,
            password: password,
        });
        if (response.status === "success") {
            this.token = response.data.token;
        }
        return response;
    }

    async newEvent(roomId: string, data: any, id: string | undefined = undefined) {
        return this.send("new_event", {
            room_id: roomId,
            events: [{
                data: data,
                id: id,
            }],
        });
    }

    async subscribe(cb: (a: EventJson) => void) {
        this.subscriptionEndpoints.push({cb: cb});
        await this._subscribe(cb);
    }

    private async _subscribe(cb: (a: EventJson) => void) {
        this.numRequests++;
        if (this.websocket !== null) {
            const subscription = await this.send('subscribe', {}, 60_000, false, (res) => {
                if (res.data && res.data.events) {
                    for (const event of res.data.events) {
                        cb(event);
                    }
                }
            });
            this.subscriptions.push(subscription);
            return subscription;
        } else {
            this.send('subscribe', {}, 60_000).then((res) => {
                if (res.status === "success") {
                    this._subscribe(cb);
                    cb(res);
                } else {
                    setTimeout(() => {
                        this._subscribe(cb);
                    }, 1000);
                }
            }).catch(() => {
                setTimeout(() => {
                    this._subscribe(cb);
                }, 1000);
            });
            return null;  // TODO subscription id
        }
    }
}
